# 
# VisicrA - Visualization InterCriteria Analysis
# 
# Author: Nikolay Ikonomov
# Version: 0.9.1
# Date: April 23, 2023
# Compiled by: Python/Tkinter/NumPy/Matplotlib
# 

# Does it show that I come from Java and C/C++/FLTK ?

# Only tkinter, ttk as last resort
# Only grid, pack as last resort
# https://tcl.tk/man/tcl8.6/TkCmd/contents.htm
# https://tk-tutorial.readthedocs.io/en/latest/entry/entry.html
# https://docs.python.org/3/library/dialog.html
# https://numpy.org/doc/stable/reference/arrays.ndarray.html
# https://mkaz.blog/working-with-python/string-formatting/
# https://matplotlib.org/stable/gallery/user_interfaces/embedding_in_tk_sgskip.html

# https://stackoverflow.com/questions/41315873/attempting-to-resolve-blurred-tkinter-text-scaling-on-windows-10-high-dpi-disp - HiDPI
# https://linuxconfig.org/how-to-build-a-tkinter-application-using-an-object-oriented-approach - thanks for the quick OOP
# https://stackoverflow.com/questions/1984104/how-to-avoid-explicit-t-in-python - thank you for saving my sanity !!!

# ~ # HiDPI
# ~ import ctypes
# ~ ctypes.windll.shcore.SetProcessDpiAwareness(1)
# ~ ws = 15
# ~ ww = 10
# ~ wx = 2

# Display 1:1
ws = 10
ww = 5
wx = 1

# Tkinter
from tkinter import *
from tkinter import ttk
from tkinter import filedialog

# Datetime
from datetime import datetime

# Regular expression
import re

# Matrix
import numpy

# Plot
import matplotlib
from matplotlib.figure import Figure
from matplotlib.ticker import FixedLocator
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.backends.backend_tkagg import NavigationToolbar2Tk

# Class access
#define cc self

# Check int
def checkInt(v) -> bool:
	try:
		int(v)
		return True
	except:
		return False

# Check float
def checkFloat(v) -> bool:
	try:
		float(v)
		return True
	except:
		return False

# Choose color
def chooseColor(arr, i, j) -> str:
	if arr[i,j] > 0.75 and arr[j,i] < 0.25:
		if i < j: return "green"
		else: return "red"
	elif arr[i,j] < 0.25 and arr[j,i] > 0.75:
		if i < j: return "red"
		else: return "green"
	else:
		return "magenta"

# Criteria matrix
def icraCriteria(crit, arr, rows, cols):
	for i in range(rows):
		n = 0
		for k in range(cols-1):
			for j in range(k, cols-1):
				crit[i,n] = numpy.sign(arr[i,k] - arr[i,j+1])
				n += 1
	return None

# Counter A - equal elements:  0==0  -1==-1  1==1  (flag==1)   0=/=0  -1==-1  1==1  (flag==0)
def icraCounterA(arr, a, b, cols, flag) -> float:
	n = 0
	for k in range(cols):
		if  (arr[a,k] ==  1 and arr[b,k] ==  1) or \
			(arr[a,k] == -1 and arr[b,k] == -1) or \
			(arr[a,k] ==  0 and arr[b,k] ==  0 and flag == 1):
			n += 1
	return float(1.0*n/cols)

# Counter B - different elements:  0==0  -1=/=1  1=/=-1  (flag==0)   0=/=0  -1=/=1  1=/=-1  (flag==1)
def icraCounterB(arr, a, b, cols, flag) -> float:
	n = 0
	for k in range(cols):
		if  (arr[a,k] == -1 and arr[b,k] ==  1) or \
			(arr[a,k] ==  1 and arr[b,k] == -1) or \
			(arr[a,k] ==  0 and arr[b,k] ==  0 and flag == 1):
			n += 1
	return float(1.0*n/cols)

# Result matrix
def icraResult(res, crit, rows, cols, flagA, flagB):
	for i in range(rows):
		# upper triangular - matrix MU-A
		for j in range(i+1,rows):
			res[i,j] = icraCounterA(crit, i, j, cols, flagA)
		# lower triangular - matrix NU-B
		for j in range(i):
			res[i,j] = icraCounterB(crit, i, j, cols, flagB)
		# ignore diagonal elements
	return None

# VisicrA extends Tk
class VisicrA(Tk):
	
	# Constructor
	def __init__(cc):
		
		# Initialize
		super().__init__()
		
		# >>> Root window
		cc.title("VisicrA v0.9.1")
		
		# Export flag
		cc.flagE = 0
		
		# Background
		backClr = "#E8E4E0"
		
		# >>> Content window
		content = PanedWindow(cc, orient=HORIZONTAL, sashwidth=ws, sashpad=0, sashrelief="raised", background=backClr, borderwidth=0)
		content.grid(row=0, column=0, sticky=(N,W,E,S))
		
		
		# >>> Left frame
		leftF = Frame(content, background=backClr)
		leftF.grid(row=0, column=0, sticky=(N,W,E,S))
		
		# >>> Buttons in left frame
		btnF = Frame(leftF, background=backClr)
		btnF.grid(row=0, column=0, sticky=(N,W), padx=(ww,ww), pady=(ww,ww))
		
		btnOpen = ttk.Button(btnF, text="Open File", width=15, command=cc.listenOpenFile)
		btnSave = ttk.Button(btnF, text="Save File", width=15, command=cc.listenSaveFile)
		btnClear = ttk.Button(btnF, text="Clear", width=15, command=cc.listenClear)
		
		cc.cmbV = ttk.Combobox(btnF, values=("\u03BC-biased", "Unbiased", "\u03BD-biased", "Balanced", "Weighted", \
			"Mean", "Median"), state="readonly", width=12)
		cc.cmbV.set("\u03BC-biased")
		cc.cmbH = ttk.Combobox(btnF, values=("No headers", "Headers"), state="readonly", width=12)
		cc.cmbH.set("Headers")
		cc.cmbT = ttk.Combobox(btnF, values=("No transpose", "Transpose"), state="readonly", width=12)
		cc.cmbT.set("No transpose")
		
		btnAnalysis = ttk.Button(btnF, text="Analysis", command=cc.listenAnalysis)
		
		btnOpen.grid(row=0, column=0, sticky=(N,W), padx=(0,ww), pady=(0,ww))
		btnSave.grid(row=0, column=1, sticky=(N,W), padx=(0,ww), pady=(0,ww))
		btnClear.grid(row=0, column=2, sticky=(N,W), padx=(0,ww), pady=(0,ww))
		
		cc.cmbV.grid(row=1, column=0, sticky=(N), padx=(0,ww), pady=(0,ww))
		cc.cmbH.grid(row=1, column=1, sticky=(N), padx=(0,ww), pady=(0,ww))
		cc.cmbT.grid(row=1, column=2, sticky=(N), padx=(0,ww), pady=(0,ww))
		
		btnAnalysis.grid(row=2, column=0, columnspan=3, sticky=(N,W,E,S))
		# <<< Buttons in left frame
		
		# >>> Input text in left frame
		inF = Frame(leftF, background=backClr)
		inF.grid(row=1, column=0, sticky=(N,W,E,S), padx=(ww,ww), pady=(0,ww))
		
		cc.textA = Text(inF, wrap=NONE, width=35, height=20)
		scrollAV = Scrollbar(inF, orient=VERTICAL)
		scrollAH = Scrollbar(inF, orient=HORIZONTAL)
		
		cc.textA.configure(yscrollcommand=scrollAV.set, xscrollcommand=scrollAH.set)
		scrollAV.configure(command=cc.textA.yview)
		scrollAH.configure(command=cc.textA.xview)
		
		cc.textA.grid(row=0, column=0, sticky=(N,W,E,S))
		scrollAV.grid(row=0, column=1, sticky=(N,S))
		scrollAH.grid(row=1, column=0, sticky=(W,E))
		
		inF.rowconfigure(0, weight=1)
		inF.columnconfigure(0, weight=1)
		# <<< Input text in left frame
		
		# >>> Output text in left frame
		outF = Frame(leftF, background=backClr)
		outF.grid(row=2, column=0, sticky=(N,W,E,S), padx=(ww,ww), pady=(0,ww))
		
		cc.textB = Text(outF, wrap=NONE, width=35, height=7)
		scrollBV = Scrollbar(outF, orient=VERTICAL)
		scrollBH = Scrollbar(outF, orient=HORIZONTAL)
		
		cc.textB.configure(yscrollcommand=scrollBV.set, xscrollcommand=scrollBH.set)
		scrollBV.configure(command=cc.textB.yview)
		scrollBH.configure(command=cc.textB.xview)
		
		cc.textB.grid(row=0, column=0, sticky=(N,W,E,S))
		scrollBV.grid(row=0, column=1, sticky=(N,S))
		scrollBH.grid(row=1, column=0, sticky=(W,E))
		
		outF.rowconfigure(0, weight=1)
		outF.columnconfigure(0, weight=1)
		# <<< Output text in left frame
		
		leftF.rowconfigure(0, weight=0)
		leftF.rowconfigure(1, weight=1)
		leftF.rowconfigure(2, weight=1)
		leftF.columnconfigure(0, weight=1)
		# <<< Left frame
		
		
		# >>> Center frame
		centerF = Frame(content, background=backClr)
		centerF.grid(row=0, column=0, sticky=(N,W,E,S))
		
		# >>> Buttons in center frame
		btnFF = Frame(centerF, background=backClr)
		btnFF.grid(row=0, column=0, sticky=(N,W), padx=(ww,ww), pady=(ww,ww))
		
		lblD = Label(btnFF, text="Digits", background=backClr)
		cc.spinD = ttk.Spinbox(btnFF, values=(1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16), wrap=True, state="readonly", width=6)
		cc.spinD.set(4)
		
		btnExport = ttk.Button(btnFF, text="Export", width=15, command=cc.listenExport)
		btnInfo = ttk.Button(btnFF, text="Info", width=15, command=cc.listenInfo)
		btnAbout = ttk.Button(btnFF, text="About", width=15, command=cc.listenAbout)
		
		lblD.grid(row=0, column=0, sticky=(E), padx=(0,ww))
		cc.spinD.grid(row=0, column=1, sticky=(W), padx=(0,ww))
		btnExport.grid(row=0, column=2, sticky=(N,W), padx=(0,ww))
		btnInfo.grid(row=0, column=3, sticky=(N,W), padx=(0,ww))
		btnAbout.grid(row=0, column=4, sticky=(N,W))
		# <<< Buttons in center frame
		
		# >>> Table in center frame
		cc.tableF = Frame(centerF, background=backClr)
		cc.tableF.grid(row=1, column=0, sticky=(N,W,E,S), padx=(ww,ww), pady=(ww,ww))
		# <<< Table in center frame
		
		centerF.rowconfigure(0, weight=0)
		centerF.rowconfigure(1, weight=1)
		centerF.columnconfigure(0, weight=1)
		# <<< Center frame
		
		
		# >>> Right frame
		rightF = Frame(content, background=backClr)
		rightF.grid(row=0, column=0, sticky=(N,W,E,S))
		
		# >>> Buttons in right frame
		btnFFF = Frame(rightF, background=backClr)
		btnFFF.grid(row=0, column=0, sticky=(N,W), padx=(ww,ww), pady=(ww,ww))
		
		lblR = Label(btnFFF, text="Refresh the plot by moving the separator or by resizing the application.", background=backClr)
		
		lblP = Label(btnFFF, text="View", background=backClr)
		cc.cmbP = ttk.Combobox(btnFFF, values=("1D plot", "2D plot", "3D plot"), state="readonly", width=12)
		cc.cmbP.set("3D plot")
		lblS = Label(btnFFF, text="Size of points", background=backClr)
		cc.spinS = ttk.Spinbox(btnFFF, values=(5,10,15,20,25,30,35,40,45,50), wrap=True, state="readonly", width=6)
		cc.spinS.set(20)
		
		lblH = Label(btnFFF, text="Histogram", background=backClr)
		cc.cmbHist = ttk.Combobox(btnFFF, values=("None", "Normalized", "Native"), state="readonly", width=12)
		cc.cmbHist.set("Native")
		lblN = Label(btnFFF, text="Number of bins", background=backClr)
		cc.spinN = ttk.Spinbox(btnFFF, values=(5,10,15,20,25,30,35,40,45,50), wrap=True, state="readonly", width=6)
		cc.spinN.set(10)
		
		lblA = Label(btnFFF, text="All options", background=backClr)
		btnApply = ttk.Button(btnFFF, text="Apply", width=15, command=cc.listenPlot3D)
		
		lblR.grid(row=0, column=0, columnspan=5, sticky=(N,W), padx=(0,ww), pady=(0,ww))
		
		lblP.grid(row=1, column=0, sticky=(E), padx=(0,ww), pady=(0,ww))
		cc.cmbP.grid(row=1, column=1, sticky=(W), padx=(0,ww), pady=(0,ww))
		lblS.grid(row=1, column=2, sticky=(E), padx=(0,ww), pady=(0,ww))
		cc.spinS.grid(row=1, column=3, sticky=(W), padx=(0,ww), pady=(0,ww))
		
		lblH.grid(row=2, column=0, sticky=(E), padx=(0,ww), pady=(0,ww))
		cc.cmbHist.grid(row=2, column=1, sticky=(W), padx=(0,ww), pady=(0,ww))
		lblN.grid(row=2, column=2, sticky=(E), padx=(0,ww), pady=(0,ww))
		cc.spinN.grid(row=2, column=3, sticky=(W), padx=(0,ww), pady=(0,ww))
		
		lblA.grid(row=1, column=4, sticky=(N), padx=(0,ww), pady=(0,ww))
		btnApply.grid(row=2, column=4, sticky=(N,W), padx=(0,ww), pady=(0,ww))
		# <<< Buttons in right frame
		
		# >>> Plot in right frame
		cc.visF = Frame(rightF, background=backClr)
		cc.visF.grid(row=1, column=0, sticky=(N,W,E,S))
		
		cc.visF.rowconfigure(0, weight=1)
		cc.visF.columnconfigure(0, weight=1)
		# <<< Plot in right frame
		
		rightF.rowconfigure(1, weight=1)
		rightF.columnconfigure(0, weight=1)
		# <<< Right frame
		
		
		content.add(leftF)
		content.add(centerF)
		content.add(rightF)
		
		content.rowconfigure(0, weight=1)
		content.columnconfigure(0, weight=1)
		# <<< Content window
		
		cc.rowconfigure(0, weight=1)
		cc.columnconfigure(0, weight=1)
		# <<< Main window
		
		# Example
		cc.textA.insert(END,
			"# Open file or copy/paste data here\n"
			"# Separators: tab semicolon comma\n"
			"# Numbers: 1.7\n\n"
			"x;E;F;G;H;I\n"
			"A;6;5;3;7;6\n"
			"B;7;7;8;1;3\n"
			"C;4;3;5;9;1\n"
			"D;4;5;6;7;8\n")
		cc.msg("VisicrA v0.9.1")
		
		return None
		
	# Message
	def msg(cc, m):
		cc.textB.insert(END, (datetime.now()).strftime("%H:%M:%S") + " " + m + "\n")
		cc.textB.see(END)
		return None
		
	# Message 2
	def msg2(cc, m):
		cc.textB.insert(END, m + "\n")
		cc.textB.see(END)
		return None
		
	# Open file
	def listenOpenFile(cc):
		fname = filedialog.askopenfilename(title="Open File", initialdir=".", filetypes=(("CSV", "*.csv"), ("Text", "*.txt"), ("All Files", "*.*")))
		if fname:
			f = open(fname, "r", encoding="utf-8")
			cc.textA.delete(1.0, END)
			cc.textA.insert(END, f.read())
			f.close()
			lines = cc.textA.get(1.0, END).splitlines()
			if lines and lines[0].startswith("#visicra"):
				v = re.split(" ", lines[0].strip())
				if len(v) >= 2 and checkInt(v[1]) and (int(v[1]) == 0 or int(v[1]) == 1):
					cc.cmbH.current(int(v[1]))
				if len(v) >= 3 and checkInt(v[2]) and (int(v[2]) == 0 or int(v[2]) == 1):
					cc.cmbT.current(int(v[2]))
			cc.msg("Opened file " + f.name)
		return None
		
	# Save file
	def listenSaveFile(cc):
		fname = filedialog.asksaveasfilename(title="Save File", initialdir=".", filetypes=(("CSV", "*.csv"), ("All Files", "*.*")))
		if fname:
			if not fname.endswith(".csv"):
				fname += ".csv"
			f = open(fname, "w", encoding="utf-8")
			f.write("#visicra {:d} {:d}\n".format(cc.cmbH.current(), cc.cmbT.current()) + cc.textA.get(1.0, END))
			f.close()
			cc.msg("Saved file " + f.name)
		return None
		
	# Clear
	def listenClear(cc):
		cc.textA.delete(1.0, END)
		cc.textB.delete(1.0, END)
		cc.msg("VisicrA v0.9.1")
		return None
		
	# Analysis
	def listenAnalysis(cc):
		
		# Clean up
		if cc.flagE == 1:
			for i in range(2*cc.rows+2+1):
				for j in range(cc.rows+1):
					Label(cc.tableF, text="", font=("Courier", 10), relief="solid", borderwidth=0, background="#E8E4E0", width=10) \
						.grid(row=i, column=j, padx=(wx,wx), pady=(wx,wx))
						
		# Return on empty text
		if cc.textA.count(1.0, END)[0] <= 1:
			cc.msg("Not enough data")
			return None
			
		# Check rows and columns
		mlen = []
		mrows = 0
		mcols = 0
		for line in cc.textA.get(1.0, END).splitlines():
			if line and not line.startswith("#"):
				row = re.split("\t|;|,", line.strip())
				k = len(row)
				mlen.append(k)
				if mcols < k:
					mcols = k
				mrows += 1
				#print(str(k) + " " + str(row))
				
		#print("mrows " + str(mrows) + " mcols " + str(mcols))
		
		# Check columns
		for i in range(len(mlen)):
			if mcols != mlen[i]:
				cc.msg("Columns must be equal")
				return None
				
		# Headers
		h = cc.cmbH.current()
		t = cc.cmbT.current()
		cc.rhead = []
		cc.chead = []
		
		# Matrix - rows x cols
		if t == 0:
			cc.rows = mrows-h
			cc.cols = mcols-h
		else:
			cc.rows = mcols-h
			cc.cols = mrows-h
		cc.arr = numpy.zeros((cc.rows,cc.cols))
		
		# Process data
		i = 0
		for line in cc.textA.get(1.0, END).splitlines():
			if line and not line.startswith("#"):
				if i == 0 and h == 1:
					cc.chead = re.split("\t|;|,", line.strip())
					cc.chead.pop(0)
				else:
					row = re.split("\t|;|,", line.strip())
					if h == 1:
						cc.rhead.append(row[0])
					for j in range(h, mcols):
						if checkFloat(row[j]):
							if t == 0:
								cc.arr[i-h,j-h] = float(row[j])
							else:
								cc.arr[j-h,i-h] = float(row[j])
						else:
							cc.msg2("not-a-number:'{}' row:{:d} col:{:d}".format(row[j], i-h+1, j-h+1))
				i += 1
				
		if t == 1:
			cc.rhead = cc.chead
			
		if h == 0:
			if t == 0:
				for i in range(cc.rows):
					cc.rhead.append("A{:d}".format(i+1))
				for j in range(cc.cols):
					cc.chead.append("B{:d}".format(j+1))
			else:
				for i in range(cc.rows):
					cc.rhead.append("B{:d}".format(i+1))
				for j in range(cc.cols):
					cc.chead.append("A{:d}".format(j+1))
					
		#print("matrix rows " + str(cc.rows) + " cols " + str(cc.cols))
		#print(cc.arr)
		#print("rowhead " + str(len(cc.rhead)) + " " + str(cc.rhead))
		#print("colhead " + str(len(cc.chead)) + " " + str(cc.chead))
		
		# Criteria matrix - rows x (cols*(cols-1))/2
		szcr = int( (cc.cols*(cc.cols-1))/2 )
		cc.crit = numpy.zeros((cc.rows,szcr))
		icraCriteria(cc.crit, cc.arr, cc.rows, cc.cols)
		
		# ICrA variant
		icravar = cc.cmbV.current() + 1
		
		# Flags - mubiased1 flags 1 0, unbiased2 flags 0 0, nubiased3 flags 0 1
		flagA = 0
		flagB = 0
		if icravar == 1: flagA = 1
		if icravar == 3: flagB = 1
		
		# Result - rows x rows
		cc.res = numpy.zeros((cc.rows,cc.rows))
		
		# ICrA variant - mubiased1/unbiased2/nubiased3 - direct call
		if icravar == 1 or icravar == 2 or icravar == 3:
			icraResult(cc.res, cc.crit, cc.rows, szcr, flagA, flagB)
		# Balanced4 - mean value for mubiasedR1 and nubiasedR3 - (R1{i,j} + R3{i,j})/2
		elif icravar == 4:
			R1 = numpy.zeros((cc.rows,cc.rows))
			R3 = numpy.zeros((cc.rows,cc.rows))
			icraResult(R1, cc.crit, cc.rows, szcr, 1, 0)
			icraResult(R3, cc.crit, cc.rows, szcr, 0, 1)
			for i in range(cc.rows):
				for j in range(cc.rows):
					if i != j:
						cc.res[i,j] = (R1[i,j] + R3[i,j])/2
		# Weighted5 - each element of unbiasedR2 divided by the sum of MU+NU - R2{i,j}/(R2{i,j} + R2{j,i})
		elif icravar == 5:
			R2 = numpy.zeros((cc.rows,cc.rows))
			icraResult(R2, cc.crit, cc.rows, szcr, 0, 0)
			for i in range(cc.rows):
				for j in range(cc.rows):
					if i != j:
						cc.res[i,j] = R2[i,j]/(R2[i,j] + R2[j,i])
		# Mean value per row subtracted from each input element
		elif icravar == 6:
			mv = numpy.mean(cc.arr, axis=1)
			ms = cc.arr - mv[:,numpy.newaxis]
			mcrit = numpy.sign(ms)
			icraResult(cc.res, mcrit, cc.rows, cc.cols, 1, 0)
		# Median value per row subtracted from each input element
		elif icravar == 7:
			mv = numpy.median(cc.arr, axis=1)
			ms = cc.arr - mv[:,numpy.newaxis]
			mcrit = numpy.sign(ms)
			icraResult(cc.res, mcrit, cc.rows, cc.cols, 1, 0)
			
		# cc.arr - rows x cols - input matrix
		# cc.crit - rows x (cols*(cols-1))/2 - criteria matrix - Hamming distance
		# cc.res - rows x rows - result matrix
		
		# >>> Table 1
		# Column header
		Label(cc.tableF, text="\u03BC", font=("Courier", 10), relief="solid", borderwidth=0, background="white", width=10) \
				.grid(row=0, column=0, padx=(wx,wx), pady=(wx,wx))
		for j in range(cc.rows):
			Label(cc.tableF, text=cc.rhead[j], font=("Courier", 10), relief="solid", borderwidth=0, background="white", width=10) \
				.grid(row=0, column=j+1, padx=(wx,wx), pady=(wx,wx))
		# Data
		for i in range(cc.rows):
			# Row header
			Label(cc.tableF, text=cc.rhead[i], font=("Courier", 10), relief="solid", borderwidth=0, background="white", width=10) \
				.grid(row=i+1, column=0, padx=(wx,wx), pady=(wx,wx))
			# Main diagonal
			Label(cc.tableF, text="---", font=("Courier", 10), relief="solid", borderwidth=0, background="white", width=10) \
				.grid(row=i+1, column=i+1, padx=(wx,wx), pady=(wx,wx))
			# Upper triangular
			for j in range(i+1,cc.rows):
				v = "{:10.{}f}".format(cc.res[i,j], cc.spinD.get())
				Label(cc.tableF, text=v, font=("Courier", 10), relief="solid", borderwidth=0, background="white", width=10, \
					foreground=chooseColor(cc.res,i,j)).grid(row=i+1, column=j+1, padx=(wx,wx), pady=(wx,wx))
			# Lower triangular
			for j in range(i):
				v = "{:10.{}f}".format(cc.res[j,i], cc.spinD.get())
				Label(cc.tableF, text=v, font=("Courier", 10), relief="solid", borderwidth=0, background="white", width=10, \
					foreground=chooseColor(cc.res,i,j)).grid(row=i+1, column=j+1, padx=(wx,wx), pady=(wx,wx))
		# <<< Table
		
		# Separator
		for j in range(cc.rows+1):
			Label(cc.tableF, text=" ", font=("Courier", 10), relief="solid", borderwidth=0, background="#E8E4E0", width=10) \
				.grid(row=cc.rows+1, column=j, padx=(wx,wx), pady=(wx,wx))
				
		# >>> Table 2
		# Column header
		Label(cc.tableF, text="\u03BD", font=("Courier", 10), relief="solid", borderwidth=0, background="white", width=10) \
				.grid(row=cc.rows+2, column=0, padx=(wx,wx), pady=(wx,wx))
		for j in range(cc.rows):
			Label(cc.tableF, text=cc.rhead[j], font=("Courier", 10), relief="solid", borderwidth=0, background="white", width=10) \
				.grid(row=cc.rows+2, column=j+1, padx=(wx,wx), pady=(wx,wx))
		# Data
		for i in range(cc.rows):
			# Row header
			Label(cc.tableF, text=cc.rhead[i], font=("Courier", 10), relief="solid", borderwidth=0, background="white", width=10) \
				.grid(row=cc.rows+2+i+1, column=0, padx=(wx,wx), pady=(wx,wx))
			# Main diagonal
			Label(cc.tableF, text="---", font=("Courier", 10), relief="solid", borderwidth=0, background="white", width=10) \
				.grid(row=cc.rows+2+i+1, column=i+1, padx=(wx,wx), pady=(wx,wx))
			# Upper triangular
			for j in range(i+1,cc.rows):
				v = "{:10.{}f}".format(cc.res[j,i], cc.spinD.get())
				Label(cc.tableF, text=v, font=("Courier", 10), relief="solid", borderwidth=0, background="white", width=10, \
					foreground=chooseColor(cc.res,i,j)).grid(row=cc.rows+2+i+1, column=j+1, padx=(wx,wx), pady=(wx,wx))
			# Lower triangular
			for j in range(i):
				v = "{:10.{}f}".format(cc.res[i,j], cc.spinD.get())
				Label(cc.tableF, text=v, font=("Courier", 10), relief="solid", borderwidth=0, background="white", width=10, \
					foreground=chooseColor(cc.res,i,j)).grid(row=cc.rows+2+i+1, column=j+1, padx=(wx,wx), pady=(wx,wx))
		# <<< Table 2
		
		# Set flag for export and display clean up
		cc.flagE = 1
		cc.msg(cc.cmbV.get())
		
		# Visualize plot
		if cc.cmbP.current() == 0:
			cc.visPlot1D()
		elif cc.cmbP.current() == 1:
			cc.visPlot2D()
		else:
			cc.visPlot3D()
			
		return None
		
	# Export
	def listenExport(cc):
		if cc.flagE == 0:
			cc.msg("Nothing to export")
			return None
		fname = filedialog.asksaveasfilename(title="Export", initialdir=".", filetypes=(("CSV", "*.csv"), ("All Files", "*.*")))
		if fname:
			if not fname.endswith(".csv"):
				fname += ".csv"
			f = open(fname, "w", encoding="utf-8")
			f.write("visicra,\u03BC,\u03BD,row,col\n")
			for i in range(cc.rows):
				for j in range(i+1,cc.rows):
					f.write("{}-{},{:.{}f},{:.{}f},{:d},{:d}\n" \
						.format(cc.rhead[i], cc.rhead[j], cc.res[i,j], cc.spinD.get(), cc.res[j,i], cc.spinD.get(), i+1, j+1))
			f.close()
			cc.msg("Export saved " + f.name)
		return None
		
	# Info
	def listenInfo(cc):
		w = Toplevel()
		w.title("Information")
		t = Text(w, wrap=NONE, width=85, height=25)
		t.grid(row=0, column=0, sticky=(N,W,E,S))
		t.insert(END,
		"Visualization InterCriteria Analysis\n\n"
		">>> Left panel\n"
		"Open CSV or text file with separators tab, semicolon, comma.\n"
		"Open MS Excel/LibreOffice Calc and copy/paste the table with optional headers.\n"
		"Select Headers if header was copied.\n"
		"All data lines (except #) are loaded as a single matrix.\n\n"
		">>> Center panel\n"
		"Colors in the table and plot are determined from \u03B1=0.75 and \u03B2=0.25:\n"
		"    \u03BC > \u03B1 and \u03BD < \u03B2 - positive consonance (green),\n"
		"    \u03BC < \u03B2 and \u03BD > \u03B1 - negative consonance (red),\n"
		"    all other cases - dissonance (magenta).\n"
		"Export button saves the result in vector format in CSV file.\n"
		"Changing digits requires button Analysis again.\n\n"
		">>> Right panel\n"
		"Refresh the plot by moving the separator or by resizing the application.\n"
		"Plot icons with four arrows and magnifying glass are useful.\n"
		"Press them again to deactivate them and use mouse normally.\n"
		"Four arrows: left mouse button moves the plot around, right mouse button zooms.\n"
		"3D plot: left mouse button rotates the plot, right mouse button zooms uniformly.\n\n")
		w.rowconfigure(0, weight=1)
		w.columnconfigure(0, weight=1)
		return None
		
	# About
	def listenAbout(cc):
		w = Toplevel()
		w.title("About")
		t = Text(w, wrap=NONE, width=85, height=25)
		t.grid(row=0, column=0, sticky=(N,W,E,S))
		t.insert(END,
		"InterCriteria Analysis proposed for the first time by this article:\n"
		"  Atanassov K., D. Mavrov, V. Atanassova,\n"
		"  Intercriteria Decision Making: A New Approach for Multicriteria Decision Making,\n"
		"    Based on Index Matrices and Intuitionistic Fuzzy Sets,\n"
		"  Issues in Intuitionistic Fuzzy Sets and Generalized Nets, Vol. 11, 2014, 1-8.\n\n"
		"Determining of consonance or dissonance:\n"
		"  Atanassov K., V. Atanassova, G. Gluhchev,\n"
		"  InterCriteria Analysis: Ideas and problems,\n"
		"  Notes on Intuitionistic Fuzzy Sets, Vol. 21, 2015, No. 1, 81-88.\n\n"
		"This software application has been developed with the partial financial support\n"
		"  of the Bulgarian National Science Fund under Grant Ref. No. KP-06-N22-1/2018\n"
		"  Theoretical research and applications of InterCriteria Analysis.\n\n\n"
		"Visualization InterCriteria Analysis\n"
		"  Version: 0.9.1\n"
		"  Date: April 23, 2023\n"
		"  Compiled by: Python/Tkinter/NumPy/Matplotlib\n\n")
		w.rowconfigure(0, weight=1)
		w.columnconfigure(0, weight=1)
		return None
		
	# Plot histogram 1D with 2D bars in 2D mode
	def visPlot1D(cc):
		fig = Figure()
		gs = fig.add_gridspec(2, 2, left=0.05, right=0.99, top=0.99, bottom=0.05, \
			width_ratios=(4,1), height_ratios=(1,4), wspace=0.05, hspace=0.05)
		vis  = fig.add_subplot(gs[1,0])
		visx = fig.add_subplot(gs[0,0], sharex=vis)
		visy = fig.add_subplot(gs[1,1], sharey=vis)
		
		# Triangle
		fuzzyX = numpy.array([0,1,0,0])
		fuzzyY = numpy.array([0,0,1,0])
		vis.plot(fuzzyX, fuzzyY, c=("black"))
		
		# Points
		plotS = int( (cc.rows*cc.rows-cc.rows)/2 )
		plotX = numpy.zeros((plotS))
		plotY = numpy.zeros((plotS))
		plotC = []
		
		k = 0
		for i in range(cc.rows):
			for j in range(i+1,cc.rows):
				plotX[k] = cc.res[i,j]
				plotY[k] = cc.res[j,i]
				plotC.append(chooseColor(cc.res,i,j))
				k += 1
				
		#print(plotX)
		#print(plotY)
		
		# Scatter points
		vis.scatter(plotX, plotY, s=(int(cc.spinS.get())), c=plotC)
		
		# Bins and plot markers
		nbins = int(cc.spinN.get())
		nmark = 1.0/nbins
		
		# Histogram 1D no labels
		#visx.hist(plotX, nbins, range=(0,1), color="cyan", edgecolor="#C8C8C8", linewidth=0.5)
		#visy.hist(plotY, nbins, range=(0,1), color="orange", orientation="horizontal", edgecolor="#C8C8C8", linewidth=0.5)
		
		# Histogram 1D with labels
		xhist, xedges = numpy.histogram(plotX, nbins, range=(0,1))
		yhist, yedges = numpy.histogram(plotY, nbins, range=(0,1))
		xbc = visx.bar(xedges[:-1], xhist, width=nmark, align="edge", color="cyan", edgecolor="#C8C8C8", linewidth=0.5)
		ybc = visy.barh(yedges[:-1], yhist, height=nmark, align="edge", color="orange", edgecolor="#C8C8C8", linewidth=0.5)
		xlbl = [v if v > 0 else "" for v in xbc.datavalues]
		ylbl = [v if v > 0 else "" for v in ybc.datavalues]
		visx.bar_label(xbc, labels=xlbl, padding=-10, fontsize=10-nbins/10)
		visy.bar_label(ybc, labels=ylbl, padding=-10, fontsize=10-nbins/10)
		
		# Error for label_type="center" when moving the plot, use padding instead
		#visx.bar_label(xbc, labels=xlbl, label_type="center", fontsize=10-nbins/10)
		#visy.bar_label(ybc, labels=ylbl, label_type="center", fontsize=10-nbins/10)
		
		# Axis options
		#vis.set_aspect("equal")
		vis.tick_params(axis="both", which="major", labelsize=10-nbins/10)
		visx.tick_params(axis="both", which="major", labelsize=10-nbins/10, labelbottom=False)
		visy.tick_params(axis="both", which="major", labelsize=10-nbins/10, labelleft=False)
		
		ticks = numpy.arange(0.0,1.1,nmark)
		tickLocator = FixedLocator(ticks)
		vis.xaxis.set_major_locator(tickLocator)
		vis.yaxis.set_major_locator(tickLocator)
		
		# Grid color="#C8C8C8"
		vis.xaxis.grid(visible=True, which="major", linestyle="solid", linewidth=0.5)
		visx.xaxis.grid(visible=True, which="major", linestyle="solid", linewidth=0.5)
		vis.yaxis.grid(visible=True, which="major", linestyle="dashed", linewidth=0.5)
		visy.yaxis.grid(visible=True, which="major", linestyle="dashed", linewidth=0.5)
		
		vis.set_xlim(-0.02,1.02)
		vis.set_ylim(-0.02,1.02)
		
		#vis.set_xlabel("Degree of agreement, \u03BC")
		#vis.set_ylabel("Degree of disagreement, \u03BD")
		
		figC = FigureCanvasTkAgg(fig, cc.visF)
		figC.draw()
		
		figT = NavigationToolbar2Tk(figC, cc.visF, pack_toolbar=False)
		figT.update()
		
		figC.get_tk_widget().grid(row=0, column=0, sticky=(N,W,E,S), padx=(ww,ww), pady=(0,ww))
		figT.grid(row=1, column=0, sticky=(W,S), padx=(ww,ww), pady=(0,ww))
		
		return None
		
	# Plot histogram 1D with 2D bars in 3D mode
	def visPlot2D(cc):
		fig = Figure()
		fig.subplots_adjust(left=0, right=1, top=1, bottom=0)
		vis = fig.add_subplot(projection="3d")
		
		# Triangle
		fuzzyX = numpy.array([0,1,0,0])
		fuzzyY = numpy.array([0,0,1,0])
		vis.plot(fuzzyX, fuzzyY, c=("black"))
		
		# Points
		plotS = int( (cc.rows*cc.rows-cc.rows)/2 )
		plotX = numpy.zeros((plotS))
		plotY = numpy.zeros((plotS))
		plotC = []
		
		k = 0
		for i in range(cc.rows):
			for j in range(i+1,cc.rows):
				plotX[k] = cc.res[i,j]
				plotY[k] = cc.res[j,i]
				plotC.append(chooseColor(cc.res,i,j))
				k += 1
				
		#print(plotX)
		#print(plotY)
		
		# Scatter points
		vis.scatter(plotX, plotY, s=(int(cc.spinS.get())), c=plotC)
		
		# Bins and plot markers
		nbins = int(cc.spinN.get())
		nmark = 1.0/nbins
		
		# Histogram 1D
		xhist, xedges = numpy.histogram(plotX, nbins, range=(0,1))
		yhist, yedges = numpy.histogram(plotY, nbins, range=(0,1))
		xhist = xhist.astype(float)
		yhist = yhist.astype(float)
		
		# Max height
		maxh = 0.0
		for i in range(len(xhist)):
			if  maxh < xhist[i]:
				maxh = xhist[i]
		for i in range(len(yhist)):
			if  maxh < yhist[i]:
				maxh = yhist[i]
		#print(maxh)
		
		# Information
		for i in range(len(xhist)):
			if xhist[i] > 0:
				cc.msg2("xhist:{:d} x:{:.{}f}:{:.{}f}".format(int(xhist[i]), \
					xedges[i], cc.spinD.get(), xedges[i+1], cc.spinD.get()))
		for i in range(len(yhist)):
			if yhist[i] > 0:
				cc.msg2("yhist:{:d} y:{:.{}f}:{:.{}f}".format(int(yhist[i]), \
					yedges[i], cc.spinD.get(), yedges[i+1], cc.spinD.get()))
		
		# Normalize histogram bars
		nhist = cc.cmbHist.current()
		if nhist == 0 or nhist == 1:
			for i in range(len(xhist)):
				xhist[i] = xhist[i]/maxh
			for i in range(len(yhist)):
				yhist[i] = yhist[i]/maxh
			maxh = 1
		
		# Display histogram bars in 3D
		if nhist == 1 or nhist == 2:
			vis.bar(xedges[:-1], xhist, -0.01, zdir="y", width=nmark, align="edge", color="cyan", edgecolor="#C8C8C8", linewidth=0.5)
			vis.bar(yedges[:-1], yhist, -0.01, zdir="x", width=nmark, align="edge", color="orange", edgecolor="#C8C8C8", linewidth=0.5)
		
		# Axis options
		vis.axis("equal")
		vis.tick_params(axis="both", which="major", labelsize=10-nbins/10)
		#vis.tick_params(axis="both", label1On=False)
		
		ticks = numpy.arange(0.0,1.1,nmark)
		tickLocator = FixedLocator(ticks)
		vis.xaxis.set_major_locator(tickLocator)
		vis.yaxis.set_major_locator(tickLocator)
		
		ztick = numpy.arange(0.0,maxh+1,1.0)
		ztickloc = FixedLocator(ztick)
		vis.zaxis.set_major_locator(ztickloc)
		
		vis.set_xlim(0,1)
		vis.set_ylim(0,1)
		vis.set_zlim(0,maxh)
		
		vis.set_xlabel("Degree of agreement, \u03BC")
		vis.set_ylabel("Degree of disagreement, \u03BD")
		vis.set_zlabel("Histogram bars")
		
		vis.view_init(elev=30, azim=45, roll=0)
		
		figC = FigureCanvasTkAgg(fig, cc.visF)
		figC.draw()
		
		figT = NavigationToolbar2Tk(figC, cc.visF, pack_toolbar=False)
		figT.update()
		
		figC.get_tk_widget().grid(row=0, column=0, sticky=(N,W,E,S), padx=(ww,ww), pady=(0,ww))
		figT.grid(row=1, column=0, sticky=(W,S), padx=(ww,ww), pady=(0,ww))
		
		return None
		
	# Plot histogram 2D with 3D bars in 3D mode
	def visPlot3D(cc):
		fig = Figure()
		fig.subplots_adjust(left=0, right=1, top=1, bottom=0)
		vis = fig.add_subplot(projection="3d")
		
		# Triangle
		fuzzyX = numpy.array([0,1,0,0])
		fuzzyY = numpy.array([0,0,1,0])
		vis.plot(fuzzyX, fuzzyY, c=("black"))
		
		# Points
		plotS = int( (cc.rows*cc.rows-cc.rows)/2 )
		plotX = numpy.zeros((plotS))
		plotY = numpy.zeros((plotS))
		plotC = []
		
		k = 0
		for i in range(cc.rows):
			for j in range(i+1,cc.rows):
				plotX[k] = cc.res[i,j]
				plotY[k] = cc.res[j,i]
				plotC.append(chooseColor(cc.res,i,j))
				k += 1
				
		#print(plotX)
		#print(plotY)
		
		# Scatter points
		vis.scatter(plotX, plotY, s=(int(cc.spinS.get())), c=plotC)
		
		# Bins and plot markers
		nbins = int(cc.spinN.get())
		nmark = 1.0/nbins
		
		# Histogram 2D
		hist, xedges, yedges = numpy.histogram2d(plotX, plotY, bins=nbins, range=[[0,1],[0,1]])
		#print(hist)
		#print(xedges)
		#print(yedges)
		
		# Mesh grid
		xp, yp = numpy.meshgrid(xedges[:-1]+0.375*nmark, yedges[:-1]+0.375*nmark, indexing="ij")
		#print(xp)
		#print(yp)
		xpos = xp.ravel()
		ypos = yp.ravel()
		zpos = 0.01
		dx = 0.2*nmark
		dy = 0.2*nmark
		dz = hist.ravel()
		
		# Max height
		maxh = 0.0
		for i in range(len(dz)):
			if  maxh < dz[i]:
				maxh = dz[i]
		#print(maxh)
		
		# Information
		for i in range(len(hist)):
			for j in range(len(hist)):
				if hist[i,j] > 0:
					cc.msg2("zhist:{:d} x:{:.{}f}:{:.{}f} y:{:.{}f}:{:.{}f}".format(int(hist[i,j]), \
						xedges[i], cc.spinD.get(), xedges[i+1], cc.spinD.get(), \
						yedges[j], cc.spinD.get(), yedges[j+1], cc.spinD.get()))
		
		# Python is epic, eh?
		#mask = dz == 0 # use with [~mask]
		mask = dz != 0 # use with [mask]
		
		# Normalize histogram bars
		nhist = cc.cmbHist.current()
		if nhist == 0 or nhist == 1:
			for i in range(len(dz)):
				dz[i] = dz[i]/maxh
			maxh = 1
		
		# Display histogram bars in 3D
		if nhist == 1 or nhist == 2:
			vis.bar3d(xpos[mask], ypos[mask], zpos, dx, dy, dz[mask], color="#BFBF00")
			#vis.bar3d(xpos, ypos, zpos, dx, dy, dz, color="#BFBF00", shade=False)
		
		# Axis options
		vis.axis("equal")
		vis.tick_params(axis="both", which="major", labelsize=10-nbins/10)
		#vis.tick_params(axis="both", label1On=False)
		
		ticks = numpy.arange(0.0,1.1,nmark)
		tickLocator = FixedLocator(ticks)
		vis.xaxis.set_major_locator(tickLocator)
		vis.yaxis.set_major_locator(tickLocator)
		
		ztick = numpy.arange(0.0,maxh+1,1.0)
		ztickloc = FixedLocator(ztick)
		vis.zaxis.set_major_locator(ztickloc)
		
		vis.set_xlim(0,1)
		vis.set_ylim(0,1)
		vis.set_zlim(0,maxh)
		
		vis.set_xlabel("Degree of agreement, \u03BC")
		vis.set_ylabel("Degree of disagreement, \u03BD")
		vis.set_zlabel("Histogram bars")
		
		vis.view_init(elev=30, azim=-135, roll=0)
		
		figC = FigureCanvasTkAgg(fig, cc.visF)
		figC.draw()
		
		figT = NavigationToolbar2Tk(figC, cc.visF, pack_toolbar=False)
		figT.update()
		
		figC.get_tk_widget().grid(row=0, column=0, sticky=(N,W,E,S), padx=(ww,ww), pady=(0,ww))
		figT.grid(row=1, column=0, sticky=(W,S), padx=(ww,ww), pady=(0,ww))
		
		return None
		
	# Listen plot 3D
	def listenPlot3D(cc):
		if cc.flagE == 1:
			if cc.cmbP.current() == 0:
				cc.visPlot1D()
			elif cc.cmbP.current() == 1:
				cc.visPlot2D()
			else:
				cc.visPlot3D()
		return None
		
# Main entry point
if __name__ == '__main__':
	app = VisicrA()
	app.mainloop()

